import numpy as np

from ..normalizations import multimoora_normalization
from ..additions import rank_preferences
from ..compromise_rankings import *
from .mcda_method import MCDA_method


class MULTIMOORA_RS(MCDA_method):
    def __init__(self):
        """Create object of the MULTIMOORA Ratio System (RS) method. This method is an integral part of the 
        MULTIMOORA method. This method is the same as the MOORA method, and it can be used 
        independently as a separate MCDA method"""
        pass


    def __call__(self, matrix, weights, types):
        """
        Score alternatives provided in decision matrix `matrix` using vector with criteria weights
        `weights` and vector with criteria types `types`.

        Parameters
        ----------
            matrix : ndarray
                Decision matrix with m alternatives in rows and n criteria in columns.
            weights: ndarray
                Criteria weights. Sum of weights must be equal to 1.
            types: ndarray
                Criteria types. Profit criteria are represented by 1 and cost by -1.

        Returns
        -------
            ndrarray
                Preference values of each alternative. The best alternative has the highest preference value. 
        """
        MULTIMOORA_RS._verify_input_data(matrix, weights, types)
        return MULTIMOORA_RS._multimoora_rs(matrix, weights, types)


    @staticmethod
    def _multimoora_rs(matrix, weights, types):
        # Normalize decision matrix using normalization method `multimoora_normalization` dedicated to the MULTIMOORA method
        norm_matrix = multimoora_normalization(matrix, types)
        # Calculate the overall performance index of alternatives as 
        # difference between sums of weighted normalized performance ratings
        # of the profit and cost criteria 
        Q_sum_profit = np.sum(weights[types == 1] * norm_matrix[:, types == 1], axis = 1)
        Q_sum_cost = np.sum(weights[types == -1] * norm_matrix[:, types == -1], axis = 1)
        Q = Q_sum_profit - Q_sum_cost
        return Q


class MULTIMOORA_RP(MCDA_method):
    def __init__(self):
        """Create object of the MULTIMOORA Reference Point (RP) method. This method is an integral part of the MULTIMOORA
        method"""
        pass


    def __call__(self, matrix, weights, types):
        """
        Score alternatives provided in decision matrix `matrix` using vector with criteria weights
        `weights` and vector with criteria types `types`.

        Parameters
        ----------
            matrix : ndarray
                Decision matrix with m alternatives in rows and n criteria in columns.
            weights: ndarray
                Criteria weights. Sum of weights must be equal to 1.
            types: ndarray
                Criteria types. Profit criteria are represented by 1 and cost by -1.

        Returns
        -------
            ndrarray
                Preference values of each alternative. The best alternative has the lowest preference value. 
        """
        MULTIMOORA_RP._verify_input_data(matrix, weights, types)
        return MULTIMOORA_RP._multimoora_rp(matrix, weights, types)


    @staticmethod
    def _multimoora_rp(matrix, weights, types):
        # Normalize decision matrix using normalization method `multimoora_normalization` dedicated to the MULTIMOORA method
        norm_matrix = multimoora_normalization(matrix, types)
    
        RR = np.zeros(norm_matrix.shape[1])

        maximums = np.amax(norm_matrix, axis = 0)
        minimums = np.amin(norm_matrix, axis = 0)
        RR[types == 1] = maximums[types == 1]
        RR[types == -1] = minimums[types == -1]
        weighted_matrix = weights * np.abs(RR - norm_matrix)
        A = np.max(weighted_matrix, axis = 1)
        return A


class MULTIMOORA_FMF(MCDA_method):
    def __init__(self):
        """Create object of the MULTIMOORA Full Multiplicative Form (FMF) method. This method is an integral part of the MULTIMOORA
        method"""
        pass


    def __call__(self, matrix, weights, types):
        """
        Score alternatives provided in decision matrix `matrix` using vector with criteria weights
        `weights` and vector with criteria types `types`.

        Parameters
        ----------
            matrix : ndarray
                Decision matrix with m alternatives in rows and n criteria in columns.
            weights: ndarray
                Criteria weights. Sum of weights must be equal to 1.
            types: ndarray
                Criteria types. Profit criteria are represented by 1 and cost by -1.

        Returns
        -------
            ndrarray
                Preference values of each alternative. The best alternative has the highest preference value. 
        """
        MULTIMOORA_FMF._verify_input_data(matrix, weights, types)
        return MULTIMOORA_FMF._multimoora_fmf(matrix, weights, types)


    @staticmethod
    def _multimoora_fmf(matrix, weights, types):
        # Normalize decision matrix using normalization method `multimoora_normalization` dedicated to the MULTIMOORA method
        norm_matrix = multimoora_normalization(matrix, types)
        A = np.prod(weights[types == 1] * norm_matrix[:, types == 1], axis = 1)
        B = np.prod(weights[types == -1] * norm_matrix[:, types == -1], axis = 1)
        U = A / B
        return U


class MULTIMOORA(MCDA_method):
    def __init__(self, compromise_rank_method = dominance_directed_graph):
        """Create object of the MULTIMOORA method.

        Parameters:
        -----------
            compromise_rank_method : function
                method determining compromise ranking based on RS, RP and FMF MULTIMOORA rankings.
                The compromise rank method is selected from `compromise_rankings`.
                It can be `borda_copeland_compromise_ranking`, `dominance_directed_graph`,
                `rank_position_method` and `improved_borda_rule` which is dedicated to the MULTIMOORA method
            
        """
        self.compromise_rank_method = compromise_rank_method


    def __call__(self, matrix, weights, types):
        """
        Score alternatives provided in decision matrix `matrix` using vector with criteria weights
        `weights` and vector with criteria types `types`.

        Parameters
        ----------
            matrix : ndarray
                Decision matrix with m alternatives in rows and n criteria in columns.
            weights: ndarray
                Criteria weights. Sum of weights must be equal to 1.
            types: ndarray
                Criteria types. Profit criteria are represented by 1 and cost by -1.

        Returns
        -------
            ndrarray
                Preference values of each alternative. The best alternative has the highest preference value. 
        """
        MULTIMOORA._verify_input_data(matrix, weights, types)
        return MULTIMOORA._multimoora(matrix, weights, types, self.compromise_rank_method)


    def _multimoora(matrix, weights, types, compromise_rank_method):

        # Calculate preference values and ranking of alternatives by MULTIMOORA RS
        multimoora_rs = MULTIMOORA_RS()
        pref_rs = multimoora_rs(matrix, weights, types)

        # in MULTIMOORA RS alternatives are sorted in descending order like in TOPSIS
        rank_rs = rank_preferences(pref_rs, reverse = True)

        # Calculate preference values and ranking of alternatives by MULTIMOORA RP
        multimoora_rp = MULTIMOORA_RP()
        pref_rp = multimoora_rp(matrix, weights, types)
        pref = np.hstack((pref_rs.reshape(-1,1), pref_rp.reshape(-1,1)))
        
        # in MULTIMOORA RP alternatives are sorted in ascending order like in VIKOR
        rank_rp = rank_preferences(pref_rp, reverse = False)
        rank = np.hstack((rank_rs.reshape(-1,1), rank_rp.reshape(-1,1)))

        # Calculate preference values and ranking of alternatives by MULTIMOORA FMF
        multimoora_fmf = MULTIMOORA_FMF()
        pref_fmf = multimoora_fmf(matrix, weights, types)
        pref = np.hstack((pref, pref_fmf.reshape(-1,1)))
        
        # in MULTIMOORA FMF alternatives are sorted in descending order like in TOPSIS
        rank_fmf = rank_preferences(pref_fmf, reverse = True)
        rank = np.hstack((rank, rank_fmf.reshape(-1,1)))

        if compromise_rank_method.__name__ == 'improved_borda_rule':
            mmoora_rank = compromise_rank_method(pref, rank)
        else:
            mmoora_rank = compromise_rank_method(rank)
            
        return mmoora_rank


import numpy as np
from ..normalizations import *
from .mcda_method import MCDA_method


class WASPAS(MCDA_method):
    def __init__(self, normalization_method = linear_normalization, lambda_param = 0.5):
        """
        Create the WASPAS method object and select normalization method `normalization_method`, default
        normalization method for WASPAS is `linear_normalization` and lambda parameter `lambda_param`, 
        which is set on 0.5 by default.

        Parameters
        -----------
            normalization_method : function
                method for decision matrix normalization chosen from `normalizations`

            lambda_param : float
                lambda parameter is between 0 and 1
        """

        self.normalization_method = normalization_method
        self.lambda_param = lambda_param


    def __call__(self, matrix, weights, types):
        """
        Score alternatives provided in decision matrix `matrix` with m alternatives and n criteria 
        using criteria `weights` and criteria `types`.

        Parameters
        ----------
            matrix : ndarray
                Decision matrix with m alternatives in rows and n criteria in columns.
            weights: ndarray
                Vector with criteria weights. Sum of weights must be equal to 1.
            types: ndarray
                Vector with criteria types. Profit criteria are represented by 1 and cost by -1.

        Returns
        -------
            ndrarray
                Vector with preference values of each alternative. The best alternative has the highest preference value. 
        """
        WASPAS._verify_input_data(matrix, weights, types)
        return WASPAS._waspas(matrix, weights, types, self.normalization_method, self.lambda_param)


    @staticmethod
    def _waspas(matrix, weights, types, normalization_method, lambda_param):
        # Normalize decision matrix
        norm_matrix = normalization_method(matrix, types)
        # Calculate the total relative importance of alternatives based on WSM
        Q1 = np.sum((norm_matrix * weights), axis = 1)
        # Calculate the total relative importance of alternatives based on WPM
        Q2 = np.prod((norm_matrix ** weights), axis = 1)
        # Determine the total relative importance of alternatives
        # If lambda is equal to 0, WASPAS method is transformed to WPM
        # If lambda is equal to 1, WASPAS becomes WSM
        Q = lambda_param * Q1 + (1 - lambda_param) * Q2
        return Q



import numpy as np
from .mcda_method import MCDA_method

class SPOTIS(MCDA_method):
    def __init__(self):
        """Create SPOTIS method object.
        """
        pass


    def __call__(self, matrix, weights, types, bounds):
        """Score alternatives provided in decision matrix `matrix` using criteria `weights` and criteria `types`.

        Parameters
        ----------
            matrix : ndarray
                Decision matrix with m alternatives in rows and n criteria in columns.
            weights: ndarray
                Vector with criteria weights. Sum of weights must be equal to 1.
            types: ndarray
                Vector with criteria types. Profit criteria are represented by 1 and cost by -1.
            bounds: ndarray
                Bounds is ndarray with 2 rows and number of columns equal to criteria number. 
                Bounds contain minimum values in the first row and maximum values in the second row 
                for each criterion. Minimum and maximum values for the same criterion cannot be 
                the same.

        Returns
        -------
            ndrarray
                Vector with preference values of each alternative. The best alternative has the lowest preference value. 
        """
        SPOTIS._verify_input_data(matrix, weights, types)
        return SPOTIS._spotis(matrix, weights, types, bounds)


    @staticmethod
    def _spotis(matrix, weights, types, bounds):
        # Determine Ideal Solution Point (ISP)
        isp = np.zeros(matrix.shape[1])
        isp[types == 1] = bounds[1, types == 1]
        isp[types == -1] = bounds[0, types == -1]

        # Calculate normalized distances
        norm_matrix = np.abs(matrix - isp) / np.abs(bounds[1, :] - bounds[0, :])
        # Calculate the normalized weighted average distance
        D = np.sum(weights * norm_matrix, axis = 1)
        return D



import numpy as np
import itertools

from ..normalizations import *
from ..distance_metrics import *
from .mcda_method import MCDA_method


class CODAS(MCDA_method):
    def __init__(self, normalization_method = linear_normalization, distance_metric = euclidean, tau = 0.02):
        """
        Create the CODAS method object and select normalization method `normalization_method`, default
        normalization method for CODAS is `linear_normalization`, distance metric 
        `distance_metric` selected from `distance_metrics`, which is `euclidean` by default and tau parameter `tau`, 
        which is set on 0.02 by default.

        Parameters
        -----------
            normalization_method : function
                method for decision matrix normalization chosen from `normalizations`

            distance_metric : functions
                method for calculating the distance between two vectors

            tau : float
                the threshold parameter between 0.01 to 0.05. If the difference between 
                Euclidean `euclidean` or other selected distances of two alternatives is less than tau, these two alternatives 
                are also compared by the Taxicab distance
        """
        self.normalization_method = normalization_method
        self.distance_metric = distance_metric
        self.tau = tau


    def __call__(self, matrix, weights, types):
        """
        Score alternatives provided in decision matrix `matrix` with m alternatives and n criteria 
        using criteria `weights` and criteria `types`.

        Parameters
        ----------
            matrix : ndarray
                Decision matrix with m alternatives in rows and n criteria in columns.
            weights: ndarray
                Vector with criteria weights. Sum of weights must be equal to 1.
            types: ndarray
                Vector with criteria types. Profit criteria are represented by 1 and cost by -1.

        Returns
        -------
            ndrarray
                Vector with preference values of each alternative. The best alternative has the highest preference value. 
        """
        CODAS._verify_input_data(matrix, weights, types)
        return CODAS._codas(self, matrix, weights, types, self.normalization_method, self.distance_metric)


    # psi 0.01 - 0.05 recommended range of tau (threshold parameter) value
    def _psi(self, x):
        return 1 if np.abs(x) >= self.tau else 0


    @staticmethod
    def _codas(self, matrix, weights, types, normalization_method, distance_metric):
        # Normalize matrix using linear normalization
        norm_matrix = normalization_method(matrix, types)
    
        # Multiply all rows of normalized matrix by weights
        weighted_matrix = norm_matrix * weights
        m, n = weighted_matrix.shape

        # Calculate NIS vector (anti-ideal solution)
        nis = np.min(weighted_matrix, axis=0)

        # Calculate chosen distance (for example Euclidean) and Taxicab distance from anti-ideal solution
        E = np.array([distance_metric(x, nis) for x in weighted_matrix])

        # Calculate Taxicab (Manhattan) distance
        T = np.sum(np.abs(weighted_matrix - nis), axis=1)
    
        # Construct the relative assessment matrix H
        h = np.zeros((m, m))
        for i, j in itertools.product(range(m), range(m)):
            h[i, j] = (E[i] - E[j]) + (self._psi(E[i] - E[j]) * (T[i] - T[j]))

        H = np.sum(h, axis=1)
        return H


import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pyrepo-mcda",
    version="0.0.4",
    author="Aleksandra Bączkiewicz",
    author_email="aleksandra.baczkiewicz@phd.usz.edu.pl",
    description="Package with Methods for Multi-Criteria Decision Analysis",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/energyinpython/pyrepo-mcda",
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.4",
	install_requires=['numpy', 'pandas']
)


import numpy as np

# linear normalization
def linear_normalization(matrix, types):
    """
    Normalize decision matrix using linear normalization method.

    Parameters
    ----------
        matrix : ndarray
            Decision matrix with m alternatives in rows and n criteria in columns
        types : ndarray
            Criteria types. Profit criteria are represented by 1 and cost by -1.

    Returns
    -------
        ndarray
            Normalized decision matrix
    """
    x_norm = np.zeros(np.shape(matrix))
    x_norm[:, types == 1] = matrix[:, types == 1] / (np.amax(matrix[:, types == 1], axis = 0))
    x_norm[:, types == -1] = np.amin(matrix[:, types == -1], axis = 0) / matrix[:, types == -1]
    return x_norm


# min-max normalization
def minmax_normalization(matrix, types):
    """
    Normalize decision matrix using minimum-maximum normalization method.

    Parameters
    ----------
        matrix : ndarray
            Decision matrix with m alternatives in rows and n criteria in columns
        types : ndarray
            Criteria types. Profit criteria are represented by 1 and cost by -1.

    Returns
    -------
        ndarray
            Normalized decision matrix
    """
    x_norm = np.zeros((matrix.shape[0], matrix.shape[1]))
    x_norm[:, types == 1] = (matrix[:, types == 1] - np.amin(matrix[:, types == 1], axis = 0)
                             ) / (np.amax(matrix[:, types == 1], axis = 0) - np.amin(matrix[:, types == 1], axis = 0))

    x_norm[:, types == -1] = (np.amax(matrix[:, types == -1], axis = 0) - matrix[:, types == -1]
                           ) / (np.amax(matrix[:, types == -1], axis = 0) - np.amin(matrix[:, types == -1], axis = 0))

    return x_norm


# max normalization
def max_normalization(matrix, types):
    """
    Normalize decision matrix using maximum normalization method.

    Parameters
    ----------
        matrix : ndarray
            Decision matrix with m alternatives in rows and n criteria in columns
        types : ndarray
            Criteria types. Profit criteria are represented by 1 and cost by -1.

    Returns
    -------
        ndarray
            Normalized decision matrix
    """
    maximes = np.amax(matrix, axis = 0)
    matrix = matrix / maximes
    matrix[:, types == -1] = 1 - matrix[:, types == -1]
    return matrix


# sum normalization
def sum_normalization(matrix, types):
    """
    Normalize decision matrix using sum normalization method.

    Parameters
    ----------
        matrix : ndarray
            Decision matrix with m alternatives in rows and n criteria in columns
        types : ndarray
            Criteria types. Profit criteria are represented by 1 and cost by -1.

    Returns
    -------
        ndarray
            Normalized decision matrix
    """
    x_norm = np.zeros((matrix.shape[0], matrix.shape[1]))
    x_norm[:, types == 1] = matrix[:, types == 1] / np.sum(matrix[:, types == 1], axis = 0)
    x_norm[:, types == -1] = (1 / matrix[:, types == -1]) / np.sum((1 / matrix[:, types == -1]), axis = 0)

    return x_norm


# vector normalization
def vector_normalization(matrix, types):
    """
    Normalize decision matrix using vector normalization method.

    Parameters
    ----------
        matrix : ndarray
            Decision matrix with m alternatives in rows and n criteria in columns
        types : ndarray
            Criteria types. Profit criteria are represented by 1 and cost by -1.

    Returns
    -------
        ndarray
            Normalized decision matrix
    """
    x_norm = np.zeros((matrix.shape[0], matrix.shape[1]))
    x_norm[:, types == 1] = matrix[:, types == 1] / (np.sum(matrix[:, types == 1] ** 2, axis = 0))**(0.5)
    x_norm[:, types == -1] = 1 - (matrix[:, types == -1] / (np.sum(matrix[:, types == -1] ** 2, axis = 0))**(0.5))

    return x_norm


# multimoora normalization
def multimoora_normalization(matrix, types):
    x_norm = matrix / ((np.sum(matrix ** 2, axis = 0))**(0.5))
    return x_norm
